/*------------------------------------------------------------------------------------------------------------
 DCF-77 Uhr mit 4 7-Segment Anzeigen  (Gem. Anode)

 Realisiert mit AtMega88PA
 
 Version 1.5
 
 Dietmar Schrder 2011
 www.zabex.de
 
 
 Quarz Oszillator mit 4MHz 
 

                   PC6 1  28 PC5   LDR
                S1 PD0 2  27 PC4   DCF77-Empfnger
                S2 PD1 3  26 PC3   D4 4.Stelle 1m
                S3 PD2 4  25 PC2   D3 3.Stelle 10m
                S4 PD3 5  24 PC1   D2 2.Stelle 1h
                S5 PD4 6  23 PC0   D1 1.Stelle 10h
                   VCC 7  22 GND
                   GND 8  21 AREF
                Q1 PB6 9  20 AVCC
                Q2 PB7 10 19 PB5   SCK   
                S6 PD5 11 18 PB4   MISO  
                S7 PD6 12 17 PB3   MOSI  
                DP PD7 13 16 PB2   200Hz Testausgang
     PowerOn DCF77 PB0 14 15 PB1   DCF-LED 


                   
D1..D4 schalten via NPN-PNP-Paar die Anode einer Ziffer
S1..S7 und DP schalten mit ULN2803 ber Vorwiderstand ein Segment

Spaltenanordnung:
    111 
   6   2 
   6   2
    777 
   5   3
   5   3
    444  dp

Pinbelegung der Anzeige:
Segment  Pin
1        7
2        6 
3        4
4        2
5        1
6        9
7       10
dp       5 
Anode    3 und 8

    
 Input:
  PC4      (Pin 27) DCF77-Empfnger
  PC5      (Pin 28) LDR (Helligkeit)  LDR an +5V und mit 470K gegen GND
  Aref     (PIN 21) 100nF Kondensator gegen GND
  VCC      (PIN7)   +5V
  AVCC     (PIN20)  +5V


 Output:
  PB0      (PIN 14) PON vom DCF77-Modul
  PC0      (PIN 23) Digit 1 (Links)
  PC1      (PIN 24) Digit 2 
  PC2      (PIN 25) Digit 3 
  PC3      (PIN 26) Digit 4 (Rechts)

  PD0      (Pin 2)  Segment 1
  PD1      (Pin 3)  Segment 2
  PD2      (Pin 4)  Segment 3
  PD3      (Pin 5)  Segment 4
  PD4      (Pin 6)  Segment 5
  PD5      (Pin 11) Segment 6
  PD6      (Pin 12) Segment 7
  PD7      (Pin 13) Dezimalpunkt
  
  PB1      (Pin 15) DCF-LED (spiegelt Pin27) 

  Timer1 - Zeitbasis
   luft mit 400Hz  bzw. 200Hz

Der DCF-77 Empfnger startet erst nach einer fallenden Flanke an seinem PON Eingang.
Ist der Signalempfang gestrt, leuchtet der rechte Dezimalpunkt. 
Nach 5/15/60 Minuten kommen nacheinander die Punkte links davon dazu.

In der Startphase wird die angenommene Sekunde angezeigt und in der ersten Stelle das
DCF-Signal mit oberem und unterem Balken (Segment 1 und 4) signalisiert.
Bei einwandfreiem Empfang  wird nach 2-3 Minuten dann die Uhrzeit angezeigt.

Die DCF-LED auf der Platine wird im dunkeln abgeschaltet.

Angezeigt wird eine interne Uhr, die vom DCF-Empfang bei korrektem Empfang mintlich gestellt wird.
Die Flanke jedes Sekundensignals wird zur Feinjustierung der hunderstel Sekunden benutzt.
Ohne Empfang luft die interne Uhr mit der Zeit weg (Quarz zu ungenau). Das knnen schon mal 2 Minuten pro Tag sein.
Das vom Quarz abgeleitete 200Hz Signal kann an PB2 gemessen werden.
 
*/



#define F_CPU 4000000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define U8 unsigned char
#define U16 unsigned int
#define U32 unsigned long
#define I16 int

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128



/*Zifferncode
    111        1
   6   2    32  2
   6   2    32  2  
    777       64 
   5   3    16  4
   5   3    16  4
    444       8    
*/

#define Z0 (1+2+4+8+16+32)   
#define Z1 (2+4)   
#define Z2 (1+2+8+16+64)   
#define Z3 (1+2+4+8+64)   
#define Z4 (2+4+32+64)   
#define Z5 (1+4+8+32+64)   
#define Z6 (1+4+8+16+32+64)   
#define Z7 (1+2+4)   
#define Z8 (1+2+4+8+16+32+64)   
#define Z9 (1+2+4+8+32+64)   
    
//  Kanle der Analogeingnge: (8 Bit ADC)
#define   AI_LDR 5
#define   DO_DCF_LED_ON PORTB|=BIT1
#define   DO_DCF_LED_OFF PORTB&=~BIT1
#define   DI_DCF77_BIT (PINC & BIT4)
#define   DO_DCF_ENABLE (PORTB &=~ BIT0)
#define   DO_DCF_DISABLE (PORTB |= BIT0)


/* DCF-77 Decoder */
/*
Whrend jeder Minute werden die Nummern von Minute, Stunde, Tag, Wochentag, 
Monat und Jahr BCD-kodiert durch Impulsmodulation der Sekundenmarken bertragen. 
Dieses "Telegramm" gilt jeweils fr die folgende Minute. 
Dabei entsprechen Sekundenmarken mit einer Dauer von 0,1 s der binren Null 
und solche mit einer Dauer von 0,2 s der binren Eins. Die Zuordnung der einzelnen 
Sekundenmarken auf die bertragene Zeitinformation zeigt das Kodierschema 


0.              Minutenbeginn (immer LOW)
1. - 14.        Reserviert, keine Bedeutung (neuedings: Verschlselte Wetterdaten)
15.             Reserveantenne aktiv
16.             Umstellung von Sommer- auf Winterzeit, oder umgekehrt
17.             Sommerzeit aktiv
18.             Winterzeit aktiv
19.             Ankndigung Schaltsekunde
20.             Zeitbeginn (Immer High)
21. - 27.       Minute 1, 2, 4, 8, 10, 20, 40
28.             Prfbit Minute
29. - 34.       Stunde 1, 2, 4, 8, 10, 20
35.             Prfbit Stunde
36. - 41.       Tag 1, 2, 4, 8, 10, 20
42. - 44.       Wochentag 1, 2, 4
45. - 49.       Monat 1, 2, 4, 8, 10
50. - 57.       Jahr 1, 2, 4, 8, 10, 20, 40, 80
58.             Prfbit Datum
59.             fehlt zur Erkennung des Minutenanfangs

(siehe auch http://www.heret.de/funkuhr/ptb.htm)

Die drei __Prfbits__ P1, P2 und P3 ergnzen jeweils die vorhergehenden Informationswrter 
(7 Bits fr die Minute, 6 Bits fr die Stunden und 22 Bits fr das Datum, einschlielich 
der Nummer des Wochentages) auf eine __gerade__ Anzahl von Einsen.


Die Zeitmarken Nr. 17 und 18 zeigen an, auf welches Zeitsystem sich die ausgesandte 
Zeitinformation bezieht. Bei Aussendung der MEZ wird die Sekundenmarke Nr. 18 auf 0,2 s
verlngert; die Sekundenmarke Nr. 17 hat eine Dauer von 0,1 s. Bei der Aussendung der 
MESZ ist es umgekehrt.

Vor einem bergang von MEZ nach MESZ oder zurck wird auerdem jeweils eine Stunde 
lang die Sekundenmarke Nr. 16 als verlngerte Marke (0,2 s) ausgesendet. Diese 
Verlngerung beginnt beim bergang von MEZ auf MESZ (von MESZ nach MEZ) um 
01.00.16 Uhr MEZ (2.00.16 Uhr MESZ) und endet um 01.59.16 Uhr MEZ (02.59.16 Uhr MESZ).

Die Sekundenmarke Nr. 19 kndigt eine Schaltsekunde an. Sie wird dann ebenfalls 
eine Stunde lang vor Einfhrung der Schaltsekunde als verlngerte Marke (0,2 s) 
ausgesendet. Schaltsekunden werden weltweit zum gleichen Zeitpunkt in die koordinierte 
Weltzeitskala UTC eingefhrt, vorzugsweise am Ende der letzen Stunde des 31. Dezember 
oder des 30. Juni. Dies bedeutet, da Schaltsekunden in der gesetzlichen Zeit der 
Bundesrepublik Deutschland eine Sekunde vor 1 Uhr MEZ am 1. Januar oder vor 
2 Uhr MESZ am 1. Juli eingeschoben werden. Bei einer Schaltsekunde am 1. Januar (1. Juli)
beginnt daher die Verlngerung der Sekundenmarke Nr. 19 um 00.00.19 Uhr MEZ (01.00.19 Uhr MESZ)
und endet um 00.59.19 Uhr MEZ (01.59.19 Uhr MESZ).

Beim Einfgen einer Schaltsekunde hat die zugehrige Minute eine Dauer von 61 Sekunden,
und die der Marke 01.00.00 Uhr MEZ (02.00.00 Uhr MESZ) vorhergehende 59. Sekundenmarke 
wird mit einer Dauer von 0,1 s ausgesendet. Die zur eingefgten 60. Sekunde gehrige 
Marke wird weggelassen (keine Trgerabsenkung).

*/



//Globale Variable
typedef struct {
  U8  bDcf77Empfang; //Wird bei Empfang jede Sekunde umgeschaltet
  U8  zeitUebernehmen;
  U8  bBit[60];      //die empfangenen Bits
  U8  erfolgreichdecodiert;
  U8  stoerung;      //Unplausibler Empfang

  U8  uhrHH;  //Die interne Uhr (wird von DCF-Uhr gestellt)
  U8  uhrMM;
  U8  uhrSS;

  U8  vorHH; //Die empfangene Zeit in der letzten Minute
  U8  vorMM;
  U8  vorJJ;

  U8  display[4]; //Enthlt die Bitmuster fr die 7-Segment Anzeigen
  U8  aussetzen;
  U8  sehrDunkel; //TRUE: LEDs werden nur wenige s eingeschaltet
}Tglob;

Tglob volatile g;

U8 ziffern[10]={Z0,Z1,Z2,Z3,Z4,Z5,Z6,Z7,Z8,Z9};

//-----------------------------------------------------------------------------



//--------------------------------------------------------
U8 getAdc(U8 ch){ 
//--------------------------------------------------------
int n;
  ADMUX= (ch  & 0x07) | (1<<REFS0) | (1<<ADLAR); //Kanalwahl; Uref=AVcc; LeftAdjusted result
  asm volatile ("NOP");
  asm volatile ("NOP");
  asm volatile ("NOP");
  ADCSRA=_BV(ADEN)  //Enable 
        |_BV(ADSC)  //Start
        |_BV(ADPS2)| _BV(ADPS0);  // Division Factor=32, also 125KHz ADC-Clock
  n=100;
  while ((ADCSRA & 0x40) && n) n--; //n ist Timeout
  return ADCH; //8 Bit Ergebnis
}

//--------------------------------------------------------
ISR(TIMER0_COMPA_vect){ //4KHz -  Display multiplexing
//--------------------------------------------------------
static U8 digit=0;
static U8 bitmask=1;
static U8 aussetzen=0;
  //Timer starten, der die LEDs wieder abschaltet (Helligkeitssteuerung)
  TCCR2B = 0;  //Timer aus
  TCNT2=0;
  TIFR2=  (1<<OCF2A);  //Bestehenden Interrupt lschen
  TCCR2B= (1<<CS21);   //Timer starten. CTC-Modus, Clk/8  ==> 500KHz

  digit++;
  digit &= 3; //Digit luft von 0..3
  bitmask<<=1;
  if (bitmask > 8) bitmask=1;
  if (aussetzen){ //Im Dunkeln: Anzeige zeitweise ganz aus
     aussetzen--; 
  }else{
     aussetzen = g.aussetzen;
  }
  if (aussetzen > 3){ //Mind. 4 Durchlufe, damit jede Ziffer mal dran kommt.
    PORTD = 0;
    PORTC = 0;
  }else{ 
    if (g.sehrDunkel){
      PORTD = g.display[digit];

      PORTC = bitmask; //PC4+5 ohne Pull-Up
	  //Erzeuge einen 1.5s Puls
	  asm("push r1");	 //mehere NOPs werden leider wegoptimiert. Push/pops nicht!
	  asm("nop");	
	  asm("pop r1");	
      PORTC = 0;
      PORTD = 0;
  	}else{
      PORTD = g.display[digit];
      PORTC = bitmask; //PC4+5 ohne Pull-Up
    }
  }
}

#if 0
//------------------------------------------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){  // alle 5ms (200Hz)
//------------------------------------------------------------------------------------------------------------
  PORTB|=BIT2;
}

#else //Das Ganze handoptimiert in Assembler, da Verzgerungen das Multiplexing stren knnten.
void __attribute__ ((naked)) TIMER1_COMPA_vect(void){ 
   asm("sbi	0x05, 2");   // High auf PB2 ausgeben
   asm("reti");
}
#endif

//------------------------------------------------------------------------------------------------------------
ISR(TIMER2_COMPA_vect){  // Leuchtdauer)
//------------------------------------------------------------------------------------------------------------
  PORTD  = 0;  //Anzeige aus
  TCCR2B = 0;  //Timer aus
  TCNT2=0;
  PORTC  = 0;  //Highside Treiber aus
}


//--------------------------------------------------------
void dcfRestart(void){
//--------------------------------------------------------
  _delay_ms(50);
  DO_DCF_DISABLE;
  _delay_ms(2000);
  DO_DCF_ENABLE;
}

//--------------------------------------------------------
void displaytest(void){
//--------------------------------------------------------
U8 i;
  //Alles an
  g.display[0] = 255;
  g.display[1] = 255;
  g.display[2] = 255;
  g.display[3] = 255;
  _delay_ms(1000);

  //Alle Segmente nacheinander

  g.display[0] = 1;
  for (i=0; i<8; i++){
    _delay_ms(500);
    g.display[0] <<=1;
  }
  

  //1.Stelle durchlaufen 0..9
  g.display[0] = 0;
  g.display[1] = 0;
  g.display[2] = 0;
  g.display[3] = 0;
  for (i=0; i<10; i++){
    g.display[0] = ziffern[i];
    _delay_ms(333);
  }
  g.display[0] = 0;

  //2.Stelle durchlaufen 0..9
  for (i=0; i<10; i++){
    g.display[1] = ziffern[i];
    _delay_ms(333);
  }
  g.display[1] = 0;

  //3.Stelle durchlaufen 0..9
  for (i=0; i<10; i++){
    g.display[2] = ziffern[i];
    _delay_ms(333);
  }
  g.display[2] = 0;

  //4.Stelle durchlaufen 0..9
  for (i=0; i<10; i++){
    g.display[3] = ziffern[i];
    _delay_ms(333);
  }
  g.display[3] = 0;

  //Dezimalpunkte durchlaufen
  g.display[0] = BIT7;
  _delay_ms(500);
  g.display[0] = 0;
  g.display[1] = BIT7;
  _delay_ms(500);
  g.display[1] = 0;
  g.display[2] = BIT7;
  _delay_ms(500);
  g.display[2] = 0;
  g.display[3] = BIT7;
  _delay_ms(500);
  g.display[3] = 0;
}


//--------------------------------------------------------
void displaykurztest(void){
//--------------------------------------------------------
  //Alles an
  g.display[0] = 255;
  g.display[1] = 255;
  g.display[2] = 255;
  g.display[3] = 255;
  _delay_ms(1000);
}
//--------------------------------------------------------
int main(void){
//--------------------------------------------------------
static int x;


static I16 iTic;        //ms seit letztem Absenkungsbeginn
static U8 bSekunde=0;    //0..59

static U8 bSteigend;    //TRUE: Tiefpass bewegt sich von 0 aufwrts (Nahe Sekundenanfang)
static I16 iTiefpass;
static U8 startphase=1; 
static U8 toggle;
static U8 hh,mm,ss;
static I16 hs;
	
static U8 tag, wtag, monat, jahr;
static U8 neuesDatum=0;
static U8 ldr;
static U8 ldrRaw;
static U8 keinEmpfang;    //Zeit ohne korrekten Empfang in Minuten 
static U8 zeitumstellung; //Wird bei gesetztem DCF-Bit 16 hochgezhlt
static U8 sommerzeit;     //Wird bei gesetztem DCF-Bit 17 hochgezhlt
static U8 winterzeit;     //Wird bei gesetztem DCF-Bit 18 hochgezhlt

static U8 flankenzaehler=0;
static U8 flankenJeMinute=0;
static U8 pegelBisher;
I16 y;
U8 pulsdauer;
U8 zeitEmpfangen;
U8 ausfallzaehler=0;
//U8 lux;

U8 bChecksum;
U8 i;



U8 *ram;
  DDRB = BIT0 | BIT1 | BIT2;
  DDRC = BIT0 | BIT1 | BIT2 | BIT3 ;
  DDRD = BIT0 | BIT1 | BIT2 | BIT3 | BIT4 | BIT5 | BIT6 | BIT7; 
  DO_DCF_DISABLE;
  DO_DCF_LED_ON;

  //Timer 0 (generiert 4KHz (250s) Takt fr Display-Multiplex)
  TCCR0A = (1<<WGM01);
  TCCR0B = (1<<CS01);  //CTC-Modus, Clk/8  ==>500 KHz
  OCR0A = 124;         //==> 4KHz 
  TIMSK0 = (1<<OCIE0A); //mit Interrupts 
  TCNT0=0;

  // Timer1 (5ms Takt fr Uhr und DCF77)
  TCCR1A=0;
  TCCR1B=(1<<WGM12) | (1<<CS11);  //CTC-Modus, Clk/8  ==> 500 KHz
  OCR1A = 2499;                                      //==> 200  Hz
  TIMSK1 = (1<<OCIE1A); //mit Interrupts 


  // Timer2 (Leuchtdauer (Helligkeit))
  TCCR2A= (1<<WGM21);
  TCCR2B= (1<<CS20);  //CTC-Modus, Clk/1  ==> 4MHz
  OCR2A = 50;  //1:Dunkel 60:Hell
  TIMSK2 = (1<<OCIE2A); //mit Interrupts 


  
   //Globale Variablen initialisieren
//    memset(&g, 0, sizeof(Tglob));               Da rastet der Chip aus
  ram=(unsigned char  *)&g;
  for (x=0; x< sizeof(Tglob); ram[x++]=0);   //Mach ich's halt von Hand

  _delay_ms(1000);
  DO_DCF_LED_OFF;


  zeitEmpfangen=0;
  sei(); //Interrupts scharf
//  displaytest();
  displaykurztest();
  DO_DCF_ENABLE;

  while(1){  

    if (!(PORTB & BIT2)) continue; //Sync auf 5ms Takt
    PORTB &=~ BIT2;
    
    if (DI_DCF77_BIT && (g.aussetzen<8)){ //Signal zur Kontrolle ausgeben (ausser im dunkeln)
      DO_DCF_LED_ON;
    }else{
      DO_DCF_LED_OFF;
    }

    iTic+=5;
    if (iTic>3000){ //Kein Signal
      iTic=3000;
      DO_DCF_ENABLE;
    }
    //Eingang abfragen und filtern
    if (DI_DCF77_BIT){ //Signalabsenkung liefert High fr 100ms oder 200ms
      iTiefpass+=5;
      if (iTiefpass>200) iTiefpass=200;  //auf 200ms begrenzen
    }else{
      iTiefpass-=5;
      if (iTiefpass<0){
        iTiefpass=0;
        bSteigend=1;
      }
    }
    //Pegelwechsel zaehlen um Signalqualitaet zu beurteilen
    if (pegelBisher!=DI_DCF77_BIT){
      pegelBisher=DI_DCF77_BIT;
      if(flankenzaehler<255) flankenzaehler++;
    }	    


    
    //Minutenanfang erkennen
    //Minutenanfang erkennen
    if (iTic > 1500){ //Minutensignal: mehr als 1s keine Absenkung
      if (59!=bSekunde){
        bSekunde=59;
      	flankenJeMinute = flankenzaehler;   //Bei strungsfreiem Empfang ist der Wert 118 (= 59*2)
        flankenzaehler=0;
        if (ausfallzaehler<2) ausfallzaehler++;
      }
    }
    else
    //Signallnge bestimmen und im Bitfeld eintragen
    if (160==iTic){ //160ms  Tiefpass ist entweder +100-60=40 oder +100+60=160
      pulsdauer=iTiefpass;
      if (iTiefpass > 90){      
        g.bBit[bSekunde]=1;
      }else{
        g.bBit[bSekunde]=0;
      }
    
      if ((iTiefpass < 5) ||   ((iTiefpass > 60) && (iTiefpass <100))){ //Unplausible Signaldauer
        if (g.stoerung <4) g.stoerung++;  
      }else{
        if (g.stoerung >0) g.stoerung--;  
      }
    }
    //Sekundenanfang erkennen
    if ((60==iTiefpass) && bSteigend){ //60ms seit Pulsbeginn
      bSteigend=0;
      iTic=60;
      bSekunde++;
      if (58==bSekunde) ausfallzaehler=0;
      if (bSekunde>59){  //Jetzt ist Minutenanfang
        bSekunde=0;
        if (0==startphase){ //beim 1. Minutenanfang nach Einschalten ist die Zeit noch nicht bertragen
          g.zeitUebernehmen = 1;  
        }

        //Auswertung zum Minutenanfang
        if (g.zeitUebernehmen){
          if (g.bBit[16]){
            zeitumstellung++;
            }else{
            if (zeitumstellung) zeitumstellung--;
          }
          if (g.bBit[17]){
            if (sommerzeit < 60) sommerzeit++;
            }else{
            if (sommerzeit) sommerzeit--;
          }
          if (g.bBit[18]){
            if (winterzeit < 60) winterzeit++;
            }else{
            if (winterzeit) winterzeit--;
          }
          bChecksum=g.bBit[21]+g.bBit[22]+g.bBit[23]+g.bBit[24]+g.bBit[25]+g.bBit[26]+g.bBit[27]+g.bBit[28]; //Minute 
          if (0==(bChecksum % 2)){ //P1
            bChecksum=g.bBit[29]+g.bBit[30]+g.bBit[31]+g.bBit[32]+g.bBit[33]+g.bBit[34]+g.bBit[35];        //Stunde  
            if (0==(bChecksum % 2)){ //P2
              if (g.bBit[20]){ //Startbit; immer gesetzt
                bChecksum=0;
                for (i=36; i<59; i++){ //Checksumme ber Datum 
                  bChecksum+=g.bBit[i];
                }
                if (0==(bChecksum % 2)){ //P3
                  neuesDatum=0;
                  keinEmpfang=0;
                  tag = g.bBit[36]+g.bBit[37]*2+g.bBit[38]*4+g.bBit[39]*8+g.bBit[40]*10+g.bBit[41]*20;
                  wtag = g.bBit[42]+g.bBit[43]*2+g.bBit[44]*4;
                  monat = g.bBit[45]+g.bBit[46]*2+g.bBit[47]*4+g.bBit[48]*8;
                  jahr = g.bBit[50]+g.bBit[51]*2+g.bBit[52]*4+g.bBit[53]*8+g.bBit[54]*10+g.bBit[55]*20+g.bBit[56]*40+g.bBit[57]*80;

                  if (g.zeitUebernehmen && (118==flankenJeMinute) && (ausfallzaehler<2)){ //Nur bei ungestrtem Empfang an der Uhr drehen
                    //bernehme Zeit
                    g.zeitUebernehmen = 0;
                    hh = g.bBit[29]+g.bBit[30]*2+g.bBit[31]*4+g.bBit[32]*8+g.bBit[33]*10+g.bBit[34]*20;
                    mm = g.bBit[21]+g.bBit[22]*2+g.bBit[23]*4+g.bBit[24]*8+g.bBit[25]*10+g.bBit[26]*20+g.bBit[27]*40;
                    ss = 0;
			              //Ist diese Zeit genau eine Minute spter als vor einer Minute? (zustzlicher Plausibilittscheck)
                    g.vorMM++;
                    if (g.vorMM>59){
                      g.vorMM=0;
                      g.vorHH++;
                      if (g.vorHH >23) g.vorHH=0;
                    }
                    if ((g.vorHH == hh) 
                    && (g.vorMM == mm) 
                    && (g.vorJJ == jahr) //Beim Jahreswechsel muss die Uhr ja nicht um Mitternacht verdreht werden
                    && (jahr>10)){       //Wir haben 2011 - und mehr als 89 Jahre soll die Uhr gar nicht laufen
                      g.erfolgreichdecodiert=1;
                    }  
                    g.vorHH=hh;
                    g.vorMM=mm;
                    g.vorJJ=jahr; 
                  }
                }//P3
              }//Bit20
            }//P2
          }//P1
        }//g.zeitUebernehmen
        if (startphase) startphase--;
        for (i=0; i<60; i++){
          g.bBit[i]=0;
        }
      }//(bSekunde>59)
      g.bDcf77Empfang=!g.bDcf77Empfang;//toggle DCF LED;
    
      //Interne Uhr feintunen
      //Hundertstel Sekunden g.hs sollte auf 6 stehen , wenn die Uhr richtig luft (60ms seit steigender Flanke)
      //Nur machen, wenn Empfang OK
      if ((keinEmpfang < 2) && (0==g.stoerung)){
        if ((hs > 6) && (hs < 56)) {
          hs --; //Uhr etwas zurckdrehen
        }else{
          if (hs != 6){
            hs++; //Uhr etwas vor drehen. berlauf wird weiter unten automatisch behandelt.
          }  
        }
      }
    }//Sekundenanfang (60ms)

    //Intere Uhr weiterdrehen
    toggle = !toggle;
    if (toggle){ //10ms sind rum
      hs++; //Hundertstel Sekunden
      if (hs>99){ //1s ist rum
        hs-=100; //Nicht auf 0 setzen, um berlauf bei Uhr vordrehen (Feintuning) zu bercksichtigen
        g.uhrSS++;
        if (g.uhrSS > 59){
          g.uhrSS=0;
          g.uhrMM++;
          if (keinEmpfang < 100){ //Wird bei Empfang mit korreten Prfsummen auf 0 gesetzt	
            keinEmpfang++;
            if (0==(keinEmpfang % 8)){
              DO_DCF_DISABLE; //alle 8 Minuten ohne Empfang den Empfnger neu starten
              iTic=0; //Wenn iTic>3000 wird der Empfnger wieder gestartet
            }
          }	
          if (g.uhrMM > 59){
            g.uhrMM=0;
            g.uhrHH++;
            if (g.uhrHH > 23){
              g.uhrHH=0;
              neuesDatum=1; //Datum wird ja nicht angezeigt. Aber wenn ich das mal einbauen will, habe ich hiermit den Tagesbeginn.
            } 
            //Zeitumstellung?
            if ((zeitumstellung > 45) && (g.uhrHH >=2) && (g.uhrHH <=3)){ //"zeitumstellung" wurde in der Stunde vor der Umstellung jede Minute hochgezhlt
              if (sommerzeit > winterzeit){ //MESZ->MEZ
                g.uhrHH--;
              }else{ //MEZ->MESZ
                g.uhrHH++;
              }
            }
          }//g.uhrMM>59
        }//g.uhrSS>59
        if (g.erfolgreichdecodiert){ //Zeit bernehmen (wenn vorhanden)
          g.erfolgreichdecodiert=0;
          zeitEmpfangen=1;
          g.uhrHH=hh;
          g.uhrMM=mm;
          g.uhrSS=0;
		    }

        //Anzeigen der Uhrzeit
        if (g.uhrHH>9){
          g.display[0] = ziffern[g.uhrHH / 10];
        }else{
          g.display[0] = 0;
        }
        g.display[1] = ziffern[g.uhrHH % 10];
        g.display[2] = ziffern[g.uhrMM / 10];
        g.display[3] = ziffern[g.uhrMM % 10];
      }//hs>99
      if (!zeitEmpfangen){ //DCF-Werte anzeigen, bis Zeit zum ersten Mal empfangen
        if (DI_DCF77_BIT){
          g.display[0] = BIT0; //Strich oben
        }else{
          g.display[0] = BIT3; //Strich unten
        }
        g.display[1] = 0;    //aus
        g.display[2] = ziffern[bSekunde / 10];
        g.display[3] = ziffern[bSekunde % 10];
      } //!erfolgreichdecodiert 
      ldrRaw = getAdc(AI_LDR); 
      if (ldrRaw > ldr) ldr=ldrRaw; //Sofort hell werden
      if (ldrRaw < ldr) ldr--;      //Mit Rampe dunkler werden
      if (ldr < 80) g.aussetzen =(84-ldr); else  g.aussetzen=3;
    	y=(ldr-80);
      y=y*2/3;
      if (y<1) y=1;
      if (y>116) y=116;
      OCR2A=y;
      if (ldr<3)  g.sehrDunkel=1;   //Kleine Hysterese, damit es nicht flackert
      if (ldr>10) g.sehrDunkel=0; 

    }//toggle
  
    if (g.stoerung>2)     g.display[3] |= BIT7; //Dezimalpunkt einschalten
    if (keinEmpfang > 4)  g.display[2] |= BIT7; //Dezimalpunkt einschalten
    if (keinEmpfang > 14) g.display[1] |= BIT7; //Dezimalpunkt einschalten
    if (keinEmpfang > 59) g.display[0] |= BIT7; //Dezimalpunkt einschalten

#if 0  //Helligkeit im Display anzeigen
          g.display[0] = 0;    //aus
          g.display[1] = ziffern[ldrRaw/ 100];
          g.display[2] = ziffern[(ldrRaw%100) / 10];
          g.display[3] = ziffern[ldrRaw % 10];
#endif
#if 0  //Pulsdauer im Display anzeigen
          g.display[1] = ziffern[pulsdauer/ 100];
          g.display[2] = ziffern[(pulsdauer%100) / 10];
          g.display[3] = ziffern[pulsdauer % 10];
#endif




  }//While 1
  
  return 0;
}
// EOF----------------------------------------------------------------------------------------------------------
